# Copyright 2024 Mario Graff (https://github.com/mgraffg)

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from os.path import isdir, isfile, join
from microtc.utils import tweet_iterator
from encexp.utils import Download, MODELS, EncExp_URL


def download(identifier: str, first: bool=True,
             base_url: str=EncExp_URL,
             outputdir: str=MODELS):
    """download"""
    if not isdir(outputdir):
        os.mkdir(outputdir)
    output = join(outputdir, f'{identifier}.json.gz')
    if isfile(output):
        try:
            if first:
                return next(tweet_iterator(output))
            return tweet_iterator(output)
        except Exception:
            os.unlink(output)
    Download(base_url + f'/{identifier}.json.gz', output)
    if first:
        return next(tweet_iterator(output))
    return tweet_iterator(output)


def download_TextModel(identifier: str, first: bool=True):
    """Download TextModel"""
    return download(identifier, first=first)



# Copyright 2024 Mario Graff (https://github.com/mgraffg)

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import numpy as np
from numpy.testing import assert_almost_equal
from sklearn.base import clone
from microtc.utils import tweet_iterator
# from encexp.tests.test_utils import samples
from encexp.utils import load_dataset, MODEL_LANG
from encexp.text_repr import TextModel, SeqTM, EncExpT


def test_TextModel():
    """Test TextModel"""
    tm = TextModel(lang='ja', pretrained=False)
    assert tm.token_list == [1, 2, 3]
    tm = TextModel(lang=None, pretrained=False)
    assert tm.token_list == [-2, -1, 2, 3, 4]
    tm = TextModel(token_list=[2, 1], pretrained=False)
    assert tm.token_list == [2, 1]


def test_TextModel_normalize():
    """Test TextModel token normalization"""

    tm = TextModel(pretrained=False)
    txt = tm.text_transformations('e s💁🏿 🤣🤣na')
    assert txt == '~e~s~e:💁~e:🤣~e:🤣~na~'
    tm = TextModel(norm_punc=True, del_punc=False, pretrained=False)
    txt = tm.text_transformations('es💁🏿.🤣,🤣 XXX')
    assert txt == '~es~e:💁~e:.~e:🤣~e:,~e:🤣~xxx~'


def test_TextModel_tokenize():
    """Test TextModel tokenize"""
    tm = TextModel(token_list=[-1, 1], pretrained=False)
    tokens = tm.tokenize('hola💁🏿 🤣dios')
    assert tokens == ['hola', 'e:💁', 'e:🤣', 'dios', 'q:~', 'q:h',
                      'q:o', 'q:l', 'q:a', 'q:~', 'q:~', 'q:d', 
                      'q:i', 'q:o', 'q:s', 'q:~']
    tm = TextModel(token_list=[7], q_grams_words=False, pretrained=False)
    tokens = tm.tokenize('buenos💁🏿dia colegas _url _usr')
    assert tokens == ['q:~buenos', 'q:buenos~', 'q:~dia~co',
                      'q:dia~col', 'q:ia~cole', 'q:a~coleg',
                      'q:~colega', 'q:colegas', 'q:olegas~']
    

def test_TextModel_get_params():
    """Test TextModel get_params"""
    tm = TextModel(token_list=[-1, 1], pretrained=False)
    kwargs = tm.get_params()
    assert kwargs['token_list'] == [-1, 1]


def test_TextModel_identifier():
    """test TextModel identifier"""
    import hashlib

    tm = TextModel(lang='zh', pretrained=False)
    diff = tm.identifier
    cdn = ' '.join([f'{k}={v}'
                    for k, v in [('lang', 'zh'), ('pretrained', False)]])
    _ = hashlib.md5(bytes(cdn, encoding='utf-8')).hexdigest()
    assert f'TextModel_{_}' == diff
    tm = TextModel(lang='es', pretrained=False)
    diff = tm.identifier
    cdn = ' '.join([f'{k}={v}'
                    for k, v in [('lang', 'es'), ('pretrained', False)]])
    _ = hashlib.md5(bytes(cdn, encoding='utf-8')).hexdigest()
    assert f'TextModel_{_}' == diff


def test_TextModel_pretrained():
    """test TextModel pretrained"""
    tm = TextModel(lang='es')
    assert len(tm.names) == 2**17


def test_SeqTM_TM():
    """test SeqTM based on TextModel"""
    from encexp.download import download_TextModel

    seq = SeqTM(lang='es', token_max_filter=2**13,
                pretrained=False)
    tm = TextModel(lang='es')
    voc = download_TextModel(tm.identifier)['vocabulary']
    voc['dict'] = {k: v for k, v in voc['dict'].items()
                   if k[:2] == 'q:' or '~' not in k[1:-1]}
    seq.set_vocabulary(voc)
    _ = seq.tokenize('buenos dias.?, . 😂tengan')
    assert _ == ['buenos', 'dias', 'e:.', 'e:?', 'e:,', 'e:.', 'e:😂', 'tengan']
    assert seq.pretrained
    seq = SeqTM(lang='es', token_max_filter=2**13)
    _ = seq.tokenize('buenos dias .?,')
    assert _ == ['buenos~dias', 'e:.~e:?', 'e:,']


def test_SeqTM_empty_punc():
    """Test empty punc"""

    # X, y = load_dataset(['mx', 'ar'], return_X_y=True)
    seq = SeqTM(lang='es')
    # tokens = seq.tokenize(X[0])
    # assert 'q:~e' not in tokens
    # assert 'e:' in tokens
    for lang in MODEL_LANG:
        if lang in ('ja', 'zh'):
            continue
        seq = SeqTM(lang=lang)
        assert '~e:' in seq.tokens
        assert seq.token_id['~e:'] == 'e:'


def test_EncExpT_identifier():
    """Test EncExpT identifier"""
    enc = EncExpT(lang='es')
    assert enc.identifier == 'EncExpT_c69aaba0f1b0783f273f85de6f599132'
    enc = EncExpT(lang='es', use_tqdm=False,
                  pretrained=False,
                  token_max_filter=2**14)
    assert enc.identifier == 'EncExpT_c69aaba0f1b0783f273f85de6f599132'


def test_EncExpT_tailored():
    """Test EncExpT tailored"""
    dataset = load_dataset('mx')
    D = list(tweet_iterator(dataset))
    enc = EncExpT(lang='es', pretrained=False)
    enc.tailored(D, tsv_filename='tailored.tsv',
                 min_pos=32,
                 filename='tailored.json.gz')
    assert enc.weights.shape[0] == 2**14
    assert enc.weights.shape[1] == 90
    W = enc.encode('buenos dias')
    assert  W.shape == (1, 90)
    X = enc.transform(D)
    assert X.shape == (2048, 90)


def test_EncExpT_pretrained():
    """Test EncExpT pretrained"""
    enc = EncExpT(lang='es', token_max_filter=2**13)
    X = enc.transform(['buenos dias'])
    assert X.shape == (1, 4985)
    assert len(enc.names) == 4985


def test_EncExpT_tailored_intercept():
    """Test EncExpT tailored"""
    dataset = load_dataset('mx')
    D = list(tweet_iterator(dataset))
    enc = EncExpT(lang='es', with_intercept=True,
                  pretrained=False)
    enc.tailored(D, tsv_filename='tailored.tsv',
                 min_pos=32,
                 filename='tailored_intercept.json.gz')
    assert enc.weights.shape[0] == 2**14
    assert enc.weights.shape[1] == 90
    assert enc.intercept.shape[0] == 90
    X = enc.transform(['buenos dias'])
    assert X.shape[1] == 90
    enc.with_intercept = False
    assert np.fabs(X - enc.transform(['buenos dias'])).sum() != 0
    enc.with_intercept = True
    X = enc.transform(D)
    X2 = enc.seqTM.transform(D) @ enc.weights
    X2 += enc.intercept
    assert_almost_equal(X, X2, decimal=5)
    enc.merge_encode = False
    X = enc.transform(D)
    assert_almost_equal(X, X2, decimal=5)


def test_EncExpT_tailored_add():
    """Test EncExpT tailored"""
    dataset = load_dataset('mx')
    D = list(tweet_iterator(dataset))
    enc = EncExpT(lang='es', token_max_filter=2**13)
    enc.tailored(D, min_pos=32)


def test_EncExpT_tailored_no_neg():
    """Test EncExpT tailored"""
    dataset = load_dataset('mx')
    D = [f'{text} de' for text in tweet_iterator(dataset)]
    enc = EncExpT(lang='es', token_max_filter=2**13)
    enc.tailored(D, min_pos=32)


def test_EncExpT_tailored_2cl():
    """Test EncExpT tailored"""
    X, y = load_dataset(['mx', 'ar'], return_X_y=True)
    D = [dict(text=text, klass=label) for text, label in zip(X, y)]
    enc = EncExpT(lang='es', pretrained=False,
                  with_intercept=True,
                  token_max_filter=2**13)
    enc.tailored(D, self_supervised=False, min_pos=32)
    assert enc.names.tolist() == ['ar', 'mx']
    

def test_EncExpT_norm():
    """Test EncExpT norm"""
    enc = EncExpT(lang='es',
                  distance=True,
                  token_max_filter=2**13)
    assert enc.norm.shape[0] == len(enc.names)
    X1 = enc.transform(['buenos dias'])
    enc.distance = False
    X2 = enc.transform(['buenos dias'])
    assert np.fabs(X1 - X2).sum() != 0


def test_TextModel_diac():
    """Test TextModel diac"""
    from unicodedata import normalize
    dataset = load_dataset('mx')
    D = list(tweet_iterator(dataset))
    tm = TextModel(del_diac=False, pretrained=False).fit(D)
    cdn = normalize('NFD', 'ñ')
    lst = [x for x in tm.names if cdn in x]
    assert len(lst) > 3
    cdn = normalize('NFD', 'á')
    lst = [x for x in tm.names if cdn in x]
    assert len(lst) > 3


def test_EncExpT_transform_dtype():
    """Test EncExpT transform type"""
    enc = EncExpT(lang='es',
                  token_max_filter=2**13)
    X = enc.transform(['buenos dias'])
    assert X.dtype == enc.precision


# Copyright 2025 Mario Graff (https://github.com/mgraffg)

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from os.path import isfile, join
import os
import numpy as np
from sklearn.base import clone
from microtc.utils import tweet_iterator, Counter
# from encexp.tests.test_utils import samples
from encexp.utils import load_dataset
from encexp.text_repr import SeqTM, EncExpT
from encexp.build_encexp import Dataset, EncExpDataset, Train, main, NegDataset


def test_Dataset_output_filename():
    """Test Dataset"""
    seq = SeqTM(lang='es')
    ds = Dataset(text_model=seq)
    assert ds.output_filename == join('.', f'{seq.identifier}.tsv')


def test_Dataset_process():
    """Test Dataset process"""
    
    dataset = load_dataset('mx')
    iter = list(tweet_iterator(dataset))
    for x in iter:
        x['klass'] = 'mx'
    seq = SeqTM(lang='es', token_max_filter=2**13)
    ds = Dataset(text_model=seq, self_supervised=False)
    ds.process(iter)
    data = open(ds.output_filename, encoding='utf-8').readlines()
    assert data[0][:2] == 'mx'

    iter = list(tweet_iterator(dataset))
    seq = SeqTM(lang='es', token_max_filter=2**13)
    words = [str(x)  for x in seq.names
             if x[:2] != 'q:' and x[:2] != 'e:']
    cnt = Counter()
    cnt.update(words)
    seq.set_vocabulary(cnt)
    ds = Dataset(text_model=seq)
    ds.process(iter)
    data = open(ds.output_filename, encoding='utf-8').readlines()
    assert len(data) <= len(iter)
    assert len(data[0].split('\t')) == 2
    os.unlink(ds.output_filename)


def test_Dataset_self_supervise():
    """Test Dataset self_supervise parameter"""
    X, y = load_dataset(['mx', 'ar'], return_X_y=True)
    D = [dict(text=text, klass=klass)
         for text, klass in zip(X, y)]
    seq = SeqTM(lang='es', token_max_filter=2**13)
    ds = Dataset(text_model=seq)
    ds.process(D)
    data = open(ds.output_filename, encoding='utf-8').readlines()
    os.unlink(ds.output_filename)
    cnt = Counter()
    for x in data:
        cnt.update(x.split('\t')[0].split(" "))
    assert len(cnt) > 2


def test_EncExpDataset():
    """Test EncExpDataset"""
    
    dataset = load_dataset('mx')
    seq = SeqTM(lang='es', token_max_filter=2**13)
    ds = EncExpDataset(text_model=seq)
    iter = list(tweet_iterator(dataset))
    ds.process(iter)
    data = open(ds.output_filename, encoding='utf-8').readlines()
    assert len(data) <= len(iter)
    assert len(data[0].split('\t')) == 2


def test_Train_labels():
    """Test labels"""
    
    dataset = load_dataset('mx')
    seq = SeqTM(lang='es', token_max_filter=2**13)
    ds = EncExpDataset(text_model=clone(seq))
    ds.process(tweet_iterator(dataset))
    train = Train(text_model=seq, min_pos=32,
                  filename=ds.output_filename)
    assert len(train.labels) == 90
    X, y = load_dataset(['mx', 'ar', 'es'], return_X_y=True)
    D = [dict(text=text, klass=label) for text, label in zip(X, y)]
    ds = EncExpDataset(text_model=clone(seq), self_supervised=False)
    ds.process(D)
    train = Train(text_model=seq, min_pos=32,
                  filename=ds.output_filename)
    assert len(train.labels) == 3
    assert len(train.labels_freq) == 3
    os.unlink(ds.output_filename)


def test_Train_training_set():
    """Test Train"""

    dataset = load_dataset('mx')
    seq = SeqTM(lang='es', token_max_filter=2**13)
    ds = EncExpDataset(text_model=clone(seq))
    # if not isfile(ds.output_filename):
    ds.process(tweet_iterator(dataset))
    train = Train(text_model=seq, min_pos=32,
                  filename=ds.output_filename)
    labels = train.labels
    X, y = train.training_set(labels[0])
    assert X.shape[0] == len(y) and X.shape[1] == len(seq.names)
    # cnt = np.where((X > 0).sum(axis=0).A1)[0].shape
    train = Train(text_model=seq, min_pos=32,
                  keep_unfreq=True,
                  filename=ds.output_filename)
    labels = train.labels
    X, y = train.training_set(labels[0])
    _, freq =  np.unique(y, return_counts=True)
    assert freq[0] > freq[1]
    # train.min_neg = 0
    # X, y = train.training_set(labels[0])
    # _, freq =  np.unique(y, return_counts=True)
    # assert freq[0] == freq[1]
    os.unlink(ds.output_filename)

    # cnt2 = np.where((X > 0).sum(axis=0).A1)[0].shape
    # assert cnt < cnt2


def test_Train_parameters():
    """Test Train"""
    
    dataset = load_dataset('mx')
    seq = SeqTM(lang='es', token_max_filter=2**13)
    ds = EncExpDataset(text_model=clone(seq))
    if not isfile(ds.output_filename):
        ds.process(tweet_iterator(dataset))
    train = Train(text_model=seq, min_pos=32,
                  filename=ds.output_filename)
    labels = train.labels
    params = train.parameters(labels[0])
    assert 'N' in params
    os.unlink(ds.output_filename)


def test_Train_store_model():
    """Test Train"""
    
    dataset = load_dataset('mx')
    enc = EncExpT(lang='es', token_max_filter=2**13,
                  pretrained=False)
    enc.pretrained = True
    ds = EncExpDataset(text_model=clone(enc.seqTM))
    ds.identifier = enc.identifier
    if not isfile(ds.output_filename):
        ds.process(tweet_iterator(dataset))
    train = Train(text_model=enc.seqTM, min_pos=32,
                  filename=ds.output_filename)
    train.identifier = enc.identifier
    train.store_model()
    assert isfile(f'{enc.identifier}.json.gz')
    os.unlink(f'{enc.identifier}.json.gz')
    os.unlink(ds.output_filename)    


def test_Train_2cl():
    """Test Train 2 labels"""
    
    X, y = load_dataset(['mx', 'ar'], return_X_y=True)
    D = [dict(text=text, klass=label) for text, label in zip(X, y)]
    enc = EncExpT(lang='es', token_max_filter=2**13,
                  pretrained=False)
    enc.pretrained = True
    ds = EncExpDataset(text_model=clone(enc.seqTM),
                       self_supervised=False)
    ds.identifier = enc.identifier
    ds.process(D)
    train = Train(text_model=enc.seqTM, min_pos=32,
                  filename=ds.output_filename,
                  self_supervised=False)
    train.identifier = enc.identifier
    train.store_model()
    assert isfile(f'{enc.identifier}.json.gz')
    dd = list(tweet_iterator(f'{enc.identifier}.json.gz'))
    assert len(dd) == 1
    assert dd[0]['label'] == ['ar', 'mx']
    os.unlink(f'{enc.identifier}.json.gz')
    os.unlink(ds.output_filename)


def test_seqtm_build():
    """Test SeqTM CLI"""

    class A:
        """Dummy"""

    dataset = load_dataset('mx')
    A.lang = 'es'
    A.file = [dataset]
    A.voc_size_exponent = 13
    A.n_jobs = -1
    main(A)


def test_NegDataset():
    """Test NegDataset"""
    freq = {'mx': 1000, 'ar': 100, 'es': 10}
    neg = NegDataset(500, freq)
    for k in range(510):
        neg.add(f'mx {k}', 'mx')
    assert len(neg.elements['mx']) == 390
    for k in range(110):
        neg.add(f'ar {k}', 'ar')
    assert len(neg.elements['ar']) == 100
    for k in range(20):
        neg.add(f'es {k}', 'es')
    assert neg.full
    assert len(neg.dataset()) == 500
    neg = NegDataset(500, {None: 500})
    for k in range(510):
        neg.add(f'unico {k}', None)
    assert neg.full


# Copyright 2024 Mario Graff (https://github.com/mgraffg)

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from microtc.utils import tweet_iterator
# from encexp.tests.test_utils import samples
from encexp.utils import load_dataset
from encexp.build_voc import compute_TextModel_vocabulary, compute_SeqTM_vocabulary


def test_compute_TextModel_vocabulary():
    """Compute vocabulary"""
    def iterator():
        """iterator"""
        return tweet_iterator(dataset)

    dataset = load_dataset('mx')
    data = compute_TextModel_vocabulary(dataset,
                                        pretrained=False,
                                        token_max_filter=20)
    assert len(data['vocabulary']['dict']) == 20
    data = compute_TextModel_vocabulary(iterator,
                                        pretrained=False,
                                        token_max_filter=20)
    assert len(data['vocabulary']['dict']) == 20
    assert data['vocabulary']['update_calls'] == 2048


def test_compute_SeqTM_vocabulary():
    """test SeqTM vocabulary"""
    dataset = load_dataset('mx')
    params = compute_TextModel_vocabulary(dataset,
                                          pretrained=False)
    data = compute_SeqTM_vocabulary(dataset,
                                    params,
                                    pretrained=False,
                                    token_max_filter=2**13)
    assert len(data['vocabulary']['dict']) == 2**13



# Copyright 2024 Mario Graff (https://github.com/mgraffg)

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
from dataclasses import dataclass
from typing import Iterable
from random import randint, shuffle
import gzip
import json
import os
from os.path import isfile, join, isdir
from sklearn.base import clone
from sklearn.svm import LinearSVC
from joblib import Parallel, delayed
import numpy as np
from microtc.utils import tweet_iterator, Counter
import encexp
from encexp.text_repr import SeqTM, EncExpT
from encexp.utils import progress_bar, uniform_sample
from encexp.download import download


@dataclass
class Dataset:
    """Dataset"""
    prefix: str=''
    dirname: str='.'
    text_model: SeqTM=None
    use_tqdm: bool=True
    self_supervised: bool=True

    @property
    def identifier(self):
        """identifier"""
        try:
            return self._identifier
        except AttributeError:
            self.identifier = self.text_model.identifier
        return self._identifier

    @identifier.setter
    def identifier(self, value):
        self._identifier = value

    @property
    def output_filename(self):
        """output filename"""
        try:
            return self._output_filename
        except AttributeError:
            _ = join(self.dirname, f'{self.prefix}{self.identifier}.tsv')
            self.output_filename = _
        return self._output_filename

    @output_filename.setter
    def output_filename(self, value):
        self._output_filename = value

    def process(self, iterator: Iterable=None):
        """process data"""
        tm = self.text_model
        with open(self.output_filename, 'w',
                  encoding='utf-8') as fpt:
            for text in progress_bar(iterator,
                                     use_tqdm=self.use_tqdm,
                                     desc=f'{self.output_filename}'):
                label = None
                if isinstance(text, dict):
                    label = text.get('klass', None)
                    text = tm.get_text(text)
                if isinstance(text, (list, tuple)):
                    text = " ".join([tm.text_transformations(x) for x in text])
                else:
                    text = tm.text_transformations(text)
                if label is None or self.self_supervised:
                    label = " ".join(sorted(set(tm.compute_tokens(text)[0])))
                if len(label) == 0:
                    continue
                print(f'{label}\t{text}', file=fpt)


@dataclass
class EncExpDataset(Dataset):
    """EncExp Dataset"""
    @property
    def keywords(self):
        """TextModel"""
        try:
            return self._keywords
        except AttributeError:
            words = download('keywords')[self.text_model.lang]
            cnt = Counter()
            cnt.update(words)
            self.keywords = cnt
        return self._keywords

    @keywords.setter
    def keywords(self, value):
        self._keywords = value

    def process(self, iterator = None):
        self.text_model.set_vocabulary(self.keywords)
        return super().process(iterator)


@dataclass
class Train:
    """Train"""
    text_model: SeqTM=None
    min_pos: int=512
    max_pos: int=int(2**14)
    min_neg: int=int(2**14)
    filename: str=None
    use_tqdm: bool=True
    with_intercept: bool=False
    n_jobs: int=-1
    self_supervised: bool=True
    keep_unfreq: bool=False

    @property
    def identifier(self):
        """identifier"""
        try:
            return self._identifier
        except AttributeError:
            self.identifier = self.text_model.identifier
        return self._identifier

    @identifier.setter
    def identifier(self, value):
        self._identifier = value    

    @property
    def estimator(self):
        try:
            return clone(self._estimator)
        except AttributeError:
            _ = LinearSVC(class_weight='balanced',
                          fit_intercept=self.with_intercept)
            self.estimator = _
        return self._estimator

    @estimator.setter
    def estimator(self, value):
        self._estimator = value

    @property
    def labels(self):
        """Labels"""
        if hasattr(self, '_labels'):
            return self._labels
        labels_freq = Counter()
        with open(self.filename, encoding='utf-8') as fpt:
            for line in fpt:
                line = line.strip()
                labels, text = line.split('\t')
                labels = labels.split()
                labels_freq.update(labels)
        labels = sorted([k for k, v in labels_freq.items() if v >= self.min_pos])
        self.labels = labels
        self.labels_freq = labels_freq
        if self.keep_unfreq and self.self_supervised:
            cnt = Counter()
            with open(self.filename, encoding='utf-8') as fpt:
                for line in fpt:
                    line = line.strip()
                    labels, text = line.split('\t')
                    labels = labels.split()
                    _labels_freq = [(k, labels_freq[k])
                                    for k in labels]
                    klass, _ = min(_labels_freq, key=lambda x: x[1])                    
                    cnt.update([klass])
            self.neg_freq = cnt
        return labels
    
    @property
    def neg_freq(self):
        """Frequency in the negative label"""
        return self._neg_freq
    
    @neg_freq.setter
    def neg_freq(self, value):
        self._neg_freq = value

    @labels.setter
    def labels(self, value):
        self._labels = value

    @property
    def labels_freq(self):
        """Labels frequency"""
        return self._labels_freq

    @labels_freq.setter
    def labels_freq(self, value):
        self._labels_freq = value

    def transform(self, data: list):
        """Transform"""
        model = self.text_model.model
        _ = [model[x] for x in data]
        return self.text_model.tonp(_)

    def filter_tokens(self, tokens, label):
        """Filter tokens on self-supervised learning"""
        if not self.self_supervised:
            return tokens
        return [x for x in tokens if x != label]
    
    def training_set_texts(self, label):
        """Training set texts"""
        self.text_model.disable_text_transformations = True
        tokenize = self.text_model.tokenize
        max_pos = min(self.max_pos,
                      self.labels_freq[label])
        num_neg = max(max_pos, self.min_neg)
        POS = []
        labels_freq = self.labels_freq
        if self.keep_unfreq and self.self_supervised:
            labels_freq = self.neg_freq
        labels_freq = {k: v for k, v in labels_freq.items() if k != label}
        if not self.keep_unfreq:
            labels_freq = {None: num_neg}
        NEG = NegDataset(num_neg, labels_freq)
        with open(self.filename, encoding='utf-8') as fpt:
            for line in fpt:
                if len(POS) >= max_pos and NEG.full:
                    break
                line = line.strip()
                labels, text = line.split('\t')
                labels = labels.split()
                tokens = tokenize(text)
                if label in labels:
                    if len(POS) < max_pos:
                        _ = self.filter_tokens(tokens, label)
                        POS.append(_)
                    continue
                if self.keep_unfreq:
                    labels_freq = [(k, self.labels_freq[k])
                                   for k in labels]
                    klass, _ = min(labels_freq, key=lambda x: x[1])
                else:
                    klass = None
                NEG.add(tokens, klass)
        return NEG.dataset(), POS

    def training_set(self, label):
        """Training set"""
        NEG, POS = self.training_set_texts(label)        
        if len(NEG) == 0 or len(POS) == 0:
            return None
        X = self.transform(POS + NEG)
        y = [1] * len(POS) + [-1] * len(NEG)
        return X, np.array(y)

    def parameters(self, label):
        """Parameteres"""
        _ = self.training_set(label)
        if _ is None:
            return None
        X, y = _
        m = self.estimator.fit(X, y)
        hy = m.decision_function(X)
        mask = (np.fabs(hy) >= 1) & (np.sign(hy) == y)
        coef = m.coef_[0].astype(np.float16)
        output = dict(N=len(y), coef=coef.tobytes().hex(),
                      label=label, no_sv=int(mask.sum()))
        if self.with_intercept:
            output['intercept'] = m.intercept_.astype(np.float16).tobytes().hex()
        return output

    def create_model(self):
        """Create model"""
        def inner(fname, label, add_label=None):
            if isfile(fname):
                return (fname, label)
            coef = self.parameters(label)
            if coef is None:
                return None
            with open(fname, 'w', encoding='utf-8') as fpt:
                if add_label is not None:
                    coef['label'] = [add_label, coef['label']]
                print(json.dumps(coef), file=fpt)
            return (fname, label)
        if not isdir(self.identifier):
            os.mkdir(self.identifier)
        if len(self.labels) == 2:
            args = [join(self.identifier, '0.json'), self.labels[1]]
            output = inner(*args, add_label=self.labels[0])
            if output is not None:
                return [output]
            return []            
        else:
            args = [(join(self.identifier, f'{iden}.json'), label)
                    for iden, label in enumerate(self.labels)]
        _ = progress_bar(args, use_tqdm=self.use_tqdm)
        args = Parallel(n_jobs=self.n_jobs)(delayed(inner)(fname, label)
                                            for fname, label in _)
        return [x for x in args if x is not None]

    def store_model(self):
        """Create and store model"""
        args = self.create_model()
        # if len(args) == 1:
        #     with gzip.open(f'{self.identifier}.json.gz', 'wb') as fpt:
        #         fname, _ = args[0]
        #         data = next(tweet_iterator(fname))
        #         data['label'] = self.labels
        #         fpt.write(bytes(json.dumps(data) + '\n',
        #                 encoding='utf-8'))
        # else:
        with gzip.open(f'{self.identifier}.json.gz', 'wb') as fpt:
            for fname, _ in args:
                data = next(tweet_iterator(fname))
                fpt.write(bytes(json.dumps(data) + '\n',
                        encoding='utf-8'))
        self.delete_tmps(args)

    def delete_tmps(self, args):
        """Delete the auxiliary files"""
        for fname, _ in args:
            os.unlink(fname)
        os.rmdir(self.identifier)


class NegDataset:
    """Uniform sample of the negatives"""
    def __init__(self, N: int, freq: dict):
        keys = list(freq)
        cnt = uniform_sample(N,
                             np.array([freq[x] for x in keys]))
        self.cnt = {k: v for k, v in zip(keys, cnt)}
        self.elements = {k: list() for k in keys}
        self.tot = N
        self.size = 0

    def add(self, data: str, label: str):
        """Add element"""
        cnt = self.cnt[label]
        dataset = self.elements[label]
        if len(dataset) < cnt:
            self.size += 1
            dataset.append(data)

    def dataset(self):
        """Dataset"""
        values = []
        for v in self.elements.values():
            values.extend(v)
        shuffle(values)
        return values

    @property
    def full(self):
        """Indicate whether the dataset has all the elements required"""
        return self.tot - self.size <= 0


def main(args):
    """CLI"""
    filename  = args.file[0]
    lang = args.lang
    token_max_filter = 2**args.voc_size_exponent
    n_jobs = args.n_jobs
    enc = EncExpT(lang=lang, pretrained=False,
                  token_max_filter=token_max_filter)
    enc.pretrained = True
    ds = EncExpDataset(text_model=clone(enc.seqTM))
    ds.identifier = enc.identifier
    if not isfile(ds.output_filename):
        ds.process(tweet_iterator(filename))
    train = Train(text_model=enc.seqTM,
                  filename=ds.output_filename, n_jobs=n_jobs)
    train.identifier = enc.identifier
    train.store_model()
    

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='EncExp')
    parser.add_argument('-v', '--version', action='version',
                        version=f'EncExp {encexp.__version__}')
    parser.add_argument('--lang', help='Language (ar | ca | de | en | es | fr | hi | in | it | ko | nl | pl | pt | ru | tl | tr )', type=str, default=None)
    parser.add_argument('--voc_size_exponent',
                        help='Vocabulary size express as log2',
                        dest='voc_size_exponent',
                        type=int, default=13)
    parser.add_argument('--n-jobs',
                        help='Number of jobs',
                        dest='n_jobs', type=int,
                        default=-1)
    parser.add_argument('file',
                        help='Input filename',
                        nargs=1, type=str)
    args = parser.parse_args()
    main(args)


# from encexp.utils import MODEL_LANG
# from encexp import SeqTM
# import json

# D = {}
# for lang in MODEL_LANG:
#     seq = SeqTM(lang=lang,
#                 token_max_filter=2**13)
#     words = [str(x) for x in seq.names
#              if x[:2] != 'q:']
#     qgrams = [str(x) for x in seq.names
#               if x[:2] == 'q:' and x[2] != '~' and x[-1] != '~']
#     qgrams.sort(key=lambda x: len(x), reverse=True)
#     cnt = 4978 - len(words)
#     if cnt > 0:
#         words.extend(qgrams[:cnt])
#     words = sorted(words)
#     D[lang] = words

# with open('keywords.json', 'w', encoding='utf-8') as fpt:
#     print(json.dumps(D), file=fpt)
